using System.Collections.Generic;
using Serializable=System.SerializableAttribute;

namespace MorrowindModManager {
    [System.Flags]
    public enum ArchiveInvalidationFlags : uint {
        Textures  = 0x1,
        Meshes    = 0x2,
        Sounds    = 0x4,
        Music     = 0x8,
        TreesLOD  = 0x10,
        Fonts     = 0x20,
        Menus     = 0x40,
        Misc      = 0x80,
        Base      = 0x100,
        IgnoreNormal    = 0x1000,
        BSAOnly         = 0x2000,
        MatchExtensions = 0x4000,
        Universal       = 0x8000,
        EditBSAs        = 0x10000,
        EditAllEntries  = 0x20000,
        HashGenAI       = 0x40000,
        HashWarn        = 0x80000,

        Default = ArchiveInvalidationFlags.BSAOnly|ArchiveInvalidationFlags.MatchExtensions|
                  ArchiveInvalidationFlags.Textures|ArchiveInvalidationFlags.EditBSAs|
                  ArchiveInvalidationFlags.EditAllEntries|ArchiveInvalidationFlags.HashGenAI
    }

    public struct ScriptLabel {
        public string Label;
        public int LineNo;

        public ScriptLabel(string label, int lineNo) {
            Label=label;
            LineNo=lineNo;
        }
    }

    public struct PluginLoadInfo {
        public string Plugin;
        public string Target;
        public bool LoadAfter;

        public PluginLoadInfo(string plugin, string target, bool loadAfter) {
            Plugin=plugin;
            Target=target;
            LoadAfter=loadAfter;
        }
    }

    public struct ScriptEspEdit {
        public readonly bool IsGMST;
        public readonly string Plugin;
        public readonly string EDID;
        public readonly string Value;

        public ScriptEspEdit(bool gmst, string plugin, string edid, string value) {
            IsGMST=gmst;
            Plugin=plugin;
            EDID=edid;
            Value=value;
        }
    }

    public class ScriptReturnData {
        public readonly List<string> IgnorePlugins=new List<string>();
        public readonly List<string> InstallPlugins=new List<string>();
        public bool InstallAllPlugins=true;
        public readonly List<string> IgnoreData=new List<string>();
        public readonly List<string> InstallData=new List<string>();
        public bool InstallAllData=true;
        public readonly List<PluginLoadInfo> LoadOrderList=new List<PluginLoadInfo>();
        public readonly List<ConflictData> ConflictsWith=new List<ConflictData>();
        public readonly List<ConflictData> DependsOn=new List<ConflictData>();
        public readonly List<string> RegisterBSAList=new List<string>();
        public bool CancelInstall=false;
        public readonly List<string> UncheckedPlugins=new List<string>();
        public readonly List<ScriptEspWarnAgainst> EspDeactivation=new List<ScriptEspWarnAgainst>();
        public readonly List<ScriptCopyDataFile> CopyDataFiles=new List<ScriptCopyDataFile>();
        public readonly List<ScriptCopyDataFile> CopyPlugins=new List<ScriptCopyDataFile>();
        public readonly List<INIEditInfo> INIEdits=new List<INIEditInfo>();
        public readonly List<ScriptEspEdit> EspEdits=new List<ScriptEspEdit>();
    }

    public class ScriptExecutationData {
        public PluginLoadInfo[] PluginOrder;
        public string[] UncheckedPlugins;
        public ScriptEspWarnAgainst[] EspDeactivationWarning;
        public ScriptEspEdit[] EspEdits;
    }

    [Serializable]
    public struct ConflictData {
        public ConflictLevel level;
        public string File;
        public int MinMajorVersion;
        public int MinMinorVersion;
        public int MaxMajorVersion;
        public int MaxMinorVersion;
        public string Comment;

        public static bool operator==(ConflictData cd, mmod o) {
            if(cd.File!=o.ModName) return false;
            if(cd.MaxMajorVersion!=0||cd.MaxMinorVersion!=0) {
                if(cd.MaxMajorVersion>o.MajorVersion) return false;
                if(cd.MaxMajorVersion==o.MajorVersion&&cd.MaxMinorVersion>o.MinorVersion) return false;
            }
            if(cd.MinMajorVersion!=0||cd.MinMinorVersion!=0) {
                if(cd.MinMajorVersion<o.MajorVersion) return false;
                if(cd.MinMajorVersion==o.MajorVersion&&cd.MinMinorVersion<o.MinorVersion) return false;
            }
            return true;
        }
        public static bool operator!=(ConflictData cd, mmod o) {
            return !(cd==o);
        }
        public override int GetHashCode() {
            return base.GetHashCode();
        }
        public override bool Equals(object obj) {
            if(obj is ConflictData) {
                ConflictData cd=(ConflictData)obj;
                if(File==cd.File&&MinMajorVersion==cd.MinMajorVersion&&MinMinorVersion==cd.MinMinorVersion&&
                    MaxMajorVersion==cd.MaxMajorVersion&&MaxMinorVersion==cd.MaxMinorVersion&&Comment==cd.Comment) {
                    return true;
                }
            } else if(obj is mmod) {
                return (this==(mmod)obj);
            }
            return false;
        }
    }

    public struct ScriptEspWarnAgainst {
        public string Plugin;
        public DeactiveStatus Status;

        public ScriptEspWarnAgainst(string plugin, DeactiveStatus status) {
            Plugin=plugin;
            Status=status;
        }
    }

    public struct ScriptCopyDataFile {
        public readonly string CopyFrom;
        public readonly string CopyTo;
        public ScriptCopyDataFile(string from, string to) {
            CopyFrom=from;
            CopyTo=to;
        }
    }

    public struct mmodCreationOptions {
        public string Name;
        public string Author;
        public string email;
        public string website;
        public string Description;
        public string Image;
        public int MajorVersion;
        public int MinorVersion;
        public int BuildVersion;
        public CompressionType CompressionType;
        public CompressionLevel DataFileCompresionLevel;
        public CompressionLevel mmodCompressionLevel;
        public string[] esps;
        public string[] espPaths;
        public string[] DataFiles;
        public string[] DataFilePaths;
        public string readme;
        public string script;
    }

    [Serializable]
    public class BSA {
        public readonly List<string> UsedBy=new List<string>();
        public readonly string FileName;
        public readonly string LowerFileName;

        public BSA(string fileName,string modname) {
            FileName=fileName;
            LowerFileName=FileName.ToLower();
            UsedBy.Add(modname);
        }
    }

    [Serializable]
    public class EspInfo {
        public const string UnknownOwner="Unknown";
        public const string BaseOwner="Base";
        public static string[] BaseFiles= { "morrowind.esm" };

        public readonly string FileName;
        public readonly string LowerFileName;
        public readonly string BelongsTo;
        public readonly List<string> MustLoadAfter=new List<string>();
        public readonly List<string> MustLoadBefore=new List<string>();
        public bool Active;
        public System.DateTime DateModified;
        public DeactiveStatus Deactivatable;
        public readonly mmod Parent;

        public ConflictDetector.HeaderInfo header;

        public void GetHeader() {
            header=ConflictDetector.TesFile.GetHeader("data files\\"+FileName);
        }

        public EspInfo(string fileName,mmod parent) {
            FileName=fileName;
            LowerFileName=FileName.ToLower();
            BelongsTo=parent.FileName.Substring(0,parent.FileName.Length-5);
            Parent=parent;
            GetHeader();
            Deactivatable=Settings.DefaultEspWarn;
        }
        public EspInfo(string fileName) {
            FileName=fileName;
            LowerFileName=FileName.ToLower();
            if(System.Array.IndexOf<string>(BaseFiles, LowerFileName)!=-1) {
                BelongsTo=BaseOwner;
            } else {
                BelongsTo=UnknownOwner;
            }
            Parent=null;
            Deactivatable=DeactiveStatus.Allow;
            GetHeader();
        }
    }

    [Serializable]
    public class DataFileInfo {
        public readonly string FileName;
        public readonly string LowerFileName;
        public uint CRC;
        private readonly List<string> UsedBy=new List<string>();

        public static bool operator==(DataFileInfo a, DataFileInfo b) {
            if(null==(object)a||null==(object)b) {
                if(null!=(object)a||null!=(object)b) return false; else return true;
            }
            return (a.LowerFileName==b.LowerFileName);
        }
        public static bool operator!=(DataFileInfo a, DataFileInfo b) {
            return !(a==b);
        }
        public override bool Equals(object obj) {
            if(!(obj is DataFileInfo)) return false;
            return this==(DataFileInfo)obj;
        }
        public override int GetHashCode() { return LowerFileName.GetHashCode(); }

        //I dont really want this to be here, but .NET does some stupid implicit convertion by calling ToString() it isn't
        public static implicit operator string(DataFileInfo dfi) {
            //throw new System.Exception("bbbb");
            return dfi.FileName;
        }

        public DataFileInfo(string s,uint crc) {
            FileName=s;
            LowerFileName=FileName.ToLower();
            CRC=crc;
        }
        public DataFileInfo(DataFileInfo orig) {
            FileName=orig.FileName;
            LowerFileName=orig.LowerFileName;
            CRC=orig.CRC;
        }

        public void AddOwner(mmod o) { if(!UsedBy.Contains(o.LowerFileName)) UsedBy.Add(o.LowerFileName); }
        public void RemoveOwner(mmod o) { 
            UsedBy.Remove(o.LowerFileName);
            if(UsedBy.Count==0) {
                Program.Data.DataFiles.Remove(this);
                System.IO.File.Delete("data files\\"+FileName);
            }
        }
        public string Owners { get { return string.Join(", ", UsedBy.ToArray()); } }
    }

    [Serializable]
    public class INIEditInfo {
        public readonly string Section;
        public readonly string Name;
        public string OldValue;
        public readonly string NewValue;
        public string Plugin;

        public INIEditInfo(string section, string name, string newvalue) {
            Section=section.ToLower();
            Name=name.ToLower();
            NewValue=newvalue;
        }
    }

    public class mwmmException : System.ApplicationException {
        public mwmmException(string message) : base(message) { }
    }

    public enum CompressionType : byte { SevenZip, Zip }
    public enum CompressionLevel : byte { VeryHigh, High, Medium, Low, VeryLow, None }
    public enum ConflictLevel { Active, NoConflict, MinorConflict, MajorConflict, Unusable }
    public enum DeactiveStatus { Allow, WarnAgainst, Disallow }
    public enum EspSortOrder { LoadOrder, Alpha, Active, Owner, FileSize, DateCreated }
    public enum mmodSortOrder { Name, Author, ConflictLevel, FileSize, DateCreated, DateInstalled, Group }

    public class EspListSorter : System.Collections.IComparer {
        public static EspSortOrder order=EspSortOrder.LoadOrder;
        public int Compare(object a, object b) {
            System.Windows.Forms.ListViewItem la=(System.Windows.Forms.ListViewItem)a;
            System.Windows.Forms.ListViewItem lb=(System.Windows.Forms.ListViewItem)b;
            EspInfo ea = (EspInfo)la.Tag;
            EspInfo eb = (EspInfo)lb.Tag;
            switch(order) {
            case EspSortOrder.LoadOrder:
                if(ea.DateModified==eb.DateModified) return 0;
                if(ea.DateModified>eb.DateModified) return 1; else return -1;
            case EspSortOrder.Alpha:
                return string.Compare(ea.FileName, eb.FileName);  
            case EspSortOrder.Active:
                if(la.Checked==lb.Checked) return 0;
                if(la.Checked) return -1; else return 1;
            case EspSortOrder.Owner:
                return string.Compare(ea.BelongsTo, eb.BelongsTo);
            case EspSortOrder.FileSize:
                long sizea=(new System.IO.FileInfo("data files\\"+ea.FileName)).Length;
                long sizeb=(new System.IO.FileInfo("data files\\"+eb.FileName)).Length;
                if(sizea==sizeb) return 0;
                if(sizea>sizeb) return -1; else return 1;
            case EspSortOrder.DateCreated:
                System.DateTime da=(new System.IO.FileInfo("data files\\"+ea.FileName)).CreationTime;
                System.DateTime db=(new System.IO.FileInfo("data files\\"+eb.FileName)).CreationTime;
                return System.DateTime.Compare(da, db);
            default: return 0;
            }
        }
    }
    public class mmodListSorter : System.Collections.IComparer {
        public static mmodSortOrder order=mmodSortOrder.Name;
        public int Compare(object a, object b) {
            System.Windows.Forms.ListViewItem la=(System.Windows.Forms.ListViewItem)a;
            System.Windows.Forms.ListViewItem lb=(System.Windows.Forms.ListViewItem)b;
            mmod oa = (mmod)la.Tag;
            mmod ob = (mmod)lb.Tag;
            System.DateTime da;
            System.DateTime db;
            switch(order) {
            case mmodSortOrder.Name:
                return string.Compare(oa.FileName, ob.FileName);
            case mmodSortOrder.Author:
                return string.Compare(oa.Author, ob.Author);
            case mmodSortOrder.ConflictLevel:
                if((int)oa.Conflict==(int)ob.Conflict) return 0;
                if((int)oa.Conflict>(int)ob.Conflict) return 1; else return 0;
            case mmodSortOrder.FileSize:
                long sizea=(new System.IO.FileInfo(Settings.mmodDir+oa.FileName)).Length;
                long sizeb=(new System.IO.FileInfo(Settings.mmodDir+ob.FileName)).Length;
                if(sizea==sizeb) return 0;
                if(sizea>sizeb) return -1; else return 1;
            case mmodSortOrder.DateCreated:
                //This is deliberately backward
                return System.DateTime.Compare(ob.CreationTime, oa.CreationTime);
            case mmodSortOrder.DateInstalled:
                //As is this
                da=(new System.IO.FileInfo(Settings.mmodDir+oa.FileName)).CreationTime;
                db=(new System.IO.FileInfo(Settings.mmodDir+ob.FileName)).CreationTime;
                return System.DateTime.Compare(db, da);
            case mmodSortOrder.Group:
                if(oa.group==ob.group) return 0;
                if(oa.group>ob.group) return -1; else return 1;
            default: return 0;
            }
        }
    }
    
}